/*
 * ***************************************************************************
 *
 *  FILE: unifi_config.c
 *
 *      Get/Set configuration to the UniFi driver.
 *
 * Copyright (C) 2008 by Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 *
 * ***************************************************************************
 */
#define _GNU_SOURCE     /* for getsubopt() */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <getopt.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <stddef.h>

#include "csr_types.h"
#include "csr_wifi_sme_prim.h"
#include "unifiio.h"
#include "unifi_config_lib.h"

#define INSTSIZE 256  /* must be >= IFNAMSIZ */

/* Structure for requesting parameters via unifi_cfg_get_info() */
struct unifi_cfg_get_info_req {
    enum unifi_cfg_command cmd;          /* UNIFI_CFG_GET */
    enum unifi_cfg_get subcmd;           /* What to get */
};

/* Union used to retrieve parameters returned from unifi_cfg_get_info() */
union unifi_cfg_get_info_arg {
    /* Request to driver */
    struct unifi_cfg_get_info_req req;
    /* Parameter returned from driver will be one of these... */
    CsrWifiSmeCoexInfo coex_info;
    CsrWifiSmePowerSaveLevel power_save_info;
    CsrWifiSmeHostPowerMode host_power_mode;
    char drv_inst[INSTSIZE];
    uf_cfg_ap_config_t ap_cfg;
};


/*
 * ---------------------------------------------------------------------------
 *  print_usage
 *
 *      List command line args
 *
 *  Arguments:
 *      None
 *
 *  Returns:
 *      None
 * ---------------------------------------------------------------------------
 */
static void
print_usage(void)
{
    fprintf(stderr, "Usage:\n");
    fprintf(stderr, " unifi_config --dev /dev/unifiudiN \n");
    fprintf(stderr, "           [ --verbose ] \n");
    fprintf(stderr, "           [ --show ] \n");
    fprintf(stderr, "           [ --wifioff ] \n");
    fprintf(stderr, "           [ --wifion ] \n");
    fprintf(stderr, "           [ --batteries {yes | no} ] \n");
    fprintf(stderr, "           [ --powersave {none | fast | full | auto} ] \n");
    fprintf(stderr, "           [ --filter {dhcp,arp,nbns,nbds,cups | none | all} ] \n");
    fprintf(stderr, "           [ --opt_in ] \n");
    fprintf(stderr, "           [ --core_dump ] \n");
    fprintf(stderr, "           [ --uapsd {be,bk,vi,vo | none | all} ] \n");
    fprintf(stderr, "           [ --maxsp {2,4,6 | all} ] \n");
    fprintf(stderr, "           [ --addts_ie {NN:NN:...:NN (in hex)} ] \n");
    fprintf(stderr, "           [ --addts_tid {NN (in hex)} ] \n");
    fprintf(stderr, "           [ --delts_tid {NN (in hex)} ] \n");
    fprintf(stderr, "           [ --strict_draftn {yes | no} ] \n");
    fprintf(stderr, "           [ --enable_okc {yes | no} ] \n");
    fprintf(stderr, "           [ --setap_cfg {conf filename with path} ] \n");
#ifdef CSR_ONLY_DEBUG
    /* Softmac native */
    fprintf(stderr, "     Softmac Native driver specific: \n");
    fprintf(stderr, "           [ --add_template [NN:NN:NN:NN:NN:NN] <priority> NNNN...NN (all in hex) ] \n");
    fprintf(stderr, "           [ --del_template [NN:NN:NN:NN:NN:NN] <priority> (all in hex) ] \n");
    fprintf(stderr, "           [ --del_template [00:00:00:00:00:00] DEAD (delete all templates) ] \n");
    fprintf(stderr, "           [ --set_filter <precedence> {ether,udi | none | all} NN__NN..NN (underscore for don't care) ] \n");
#endif /* CSR_ONLY_DEBUG */
    fprintf(stderr, "\n");
    fprintf(stderr, "Note that a valid ADDTS command must specify both --addts_ie and --addts_tid\n");
    fprintf(stderr, "Note that after Setting the AP configuration AP needs to be restarted for the new configuration to be applied\n");
    fprintf(stderr, "\n");
}

/*
 * ---------------------------------------------------------------------------
 *  trace_..._type
 *
 *      Display text description of constant type
 *
 *  Arguments:
 *      Constant value
 *
 *  Returns:
 *      Description string
 * ---------------------------------------------------------------------------
 */
static const char*
trace_traffic_type(const CsrWifiSmeTrafficType traffic)
{
    switch (traffic)
    {
        case CSR_WIFI_SME_TRAFFIC_TYPE_OCCASIONAL:
            return "Occasional";
        case CSR_WIFI_SME_TRAFFIC_TYPE_BURSTY:
            return "Bursty";
        case CSR_WIFI_SME_TRAFFIC_TYPE_PERIODIC:
            return "Periodic";
        case CSR_WIFI_SME_TRAFFIC_TYPE_CONTINUOUS:
            return "Continuous";
        default:
            return "ERROR: unrecognised Traffic Type";
    }
 }


static const char*
trace_power_save_type(const CsrWifiSmePowerSaveLevel power_save)
{
    switch (power_save)
    {
        case CSR_WIFI_SME_POWER_SAVE_LEVEL_LOW:
            return "Active";
        case CSR_WIFI_SME_POWER_SAVE_LEVEL_HIGH:
            return "Full";
        case CSR_WIFI_SME_POWER_SAVE_LEVEL_MED:
            return "Fast";
        case CSR_WIFI_SME_POWER_SAVE_LEVEL_AUTO:
            return "Auto";
        default:
            return "ERROR: unrecognised Power Save Mode";
    }
 }


static const char*
trace_coex_scheme_type(const CsrWifiSmeCoexScheme coex_scheme)
{
    switch (coex_scheme)
    {
        case CSR_WIFI_SME_COEX_SCHEME_DISABLED:
            return "None";
        case CSR_WIFI_SME_COEX_SCHEME_CSR:
            return "CSR";
        case CSR_WIFI_SME_COEX_SCHEME_CSR_CHANNEL:
            return "CSR Channel";
        case CSR_WIFI_SME_COEX_SCHEME_PTA:
            return "PTA";
        default:
            return "ERROR: unrecognised Coex Scheme";
    }
 }


static const char*
trace_power_supply_type(const CsrWifiSmeHostPowerMode power_supply)
{
    switch (power_supply)
    {
        case CSR_WIFI_SME_HOST_POWER_MODE_ACTIVE:
            return "Running on mains";
        case CSR_WIFI_SME_HOST_POWER_MODE_POWER_SAVE:
            return "Running on batteries";
        case CSR_WIFI_SME_HOST_POWER_MODE_FULL_POWER_SAVE:
            return "Unspecified power supply";
        default:
            return "ERROR: unrecognised power supply";
    }
 }


/*
 * ---------------------------------------------------------------------------
 *  read_byte
 *  is_delim
 *
 *      Support fns.
 *
 *  Arguments:
 *      None.
 *
 *  Returns:
 *      None.
 * ---------------------------------------------------------------------------
 */
static int
read_byte(char *p, unsigned char *b)
{
    int n = 0;
    int i, c, v;

    /* Use a loop so we accept single digit fields */
    v = 0;
    for (i = 0; i < 2; i++) {

        c = *p++;

        if ((c >= '0') && (c <= '9')) {
            c -= '0';
        } else if ((c >= 'A') && (c <= 'F')) {
            c = c - 'A' + 10;
        } else if ((c >= 'a') && (c <= 'f')) {
            c = c - 'a' + 10;
        } else {
            break;
        }

        v = (v * 16) + c;
        n++;
    }

    if (n) {
        *b = v;
    }

    return n;
} /* read_byte() */


static int
is_delim(int ch)
{
    return ((ch == ':') || (ch == '.') || (ch == ','));
} /* is_delim() */


/*
 * ---------------------------------------------------------------------------
 *  read_ie
 *
 *
 *  Arguments:
 *      arg             The command-line argument string
 *      periodic_ie     Pointer to the array to receive IE.
 *
 *  Returns:
 *      number of bytes on success, -1 if the string does not conform to expected format.
 * ---------------------------------------------------------------------------
 */
static int
read_ie(char *arg, unsigned char *periodic_ie)
{
    char *p = arg;
    int i, n, offset;

    i = 0;
    n = 0;
    offset = 0;
    do {
        n = read_byte(p, &periodic_ie[offset]);
        p += n;

        if (!is_delim(*p)) break;

        p++;
        offset ++;
    } while ((n > 0) && (n < UNIFI_CFG_MAX_ADDTS_IE_LENGTH));

    return offset+1;
} /* read_ie() */


/*
 * ---------------------------------------------------------------------------
 *  main
 *
 *      Entry point
 *
 *  Arguments:
 *      argc, argv
 *
 *  Returns:
 *      non-zero on error
 * ---------------------------------------------------------------------------
 */
int
main(int argc, char **argv)
{
    int verbose = 0;
    char device[32];
    int start;
    int index;
    int r = 0;
    int fd;
    unifi_cfg_power_t wifi_power = UNIFI_CFG_POWER_UNSPECIFIED;
    unifi_cfg_powersave_t wifi_powersave = UNIFI_CFG_POWERSAVE_UNSPECIFIED;
    unifi_cfg_powersupply_t wifi_powersupply = UNIFI_CFG_POWERSUPPLY_UNSPECIFIED;
    CsrBool strict_draftn_set = FALSE;
    CsrBool strict_draftn = FALSE;
    CsrBool enable_okc_set = FALSE;
    CsrBool enable_okc = TRUE;
    CsrBool ap_cfg = FALSE;
    unsigned char wifi_filter_set;
    unsigned int wifi_filter;
    unsigned int wifi_filter_opt_in_flag = 0;
    unsigned int fw_core_dump_flag = 0;
    unsigned char uapsd_mask_set;
    CsrUint8 uapsd_mask;
    unsigned char maxsp_set;
    CsrUint8 maxsp_length;
    CsrUint8 wmm_qos_info;
    unsigned char wmm_addts_tid_set;
    CsrUint32 wmm_addts_tid;
    unsigned char wmm_delts_tid_set;
    CsrUint32 wmm_delts_tid;
    unsigned char wmm_addts_ie_set;
    unsigned char addts_ie_length;
    CsrUint8 addts_ie[UNIFI_CFG_MAX_ADDTS_IE_LENGTH];
    unsigned char show_set;
    char file_name[UNIFI_CFG_MAX_FILE_LENGTH];
    char phymode[5];
#ifdef CSR_ONLY_DEBUG
    /* Softmac native */
    unsigned char add_template = 0;
    unsigned char del_template = 0;
    CsrUint8 *template_mac = 0;
    CsrUint16 template_priority = 0;
    CsrUint8 template_len = 0;
    CsrUint8 *template_data = 0;
    unsigned char set_filter = 0;
    CsrUint8 filter_precedence = 0;
    CsrUint8 filter_action = 0;
    CsrUint8 filter_len = 0;
    CsrUint8 *filter_content = 0;
    CsrUint8 *filter_mask = 0;
#endif /* CSR_ONLY_DEBUG */
    int option_index = 0, optc;
    char *value;
    unsigned char uchar_value;

    static const struct option long_options[] = {
               {"dev", 1, 0, 0},
               {"show", 0, 0, 0},
               {"wifion", 0, 0, 0},
               {"wifioff", 0, 0, 0},
               {"powersave", 1, 0, 0},
               {"batteries", 1, 0, 0},
               {"filter", 1, 0, 0},
               {"opt_in", 0, 0, 0},
               {"core_dump", 0, 0, 0},
               {"uapsd", 1, 0, 0},
               {"maxsp", 1, 0, 0},
               {"addts_ie", 1, 0, 0},
               {"addts_tid", 1, 0, 0},
               {"delts_tid", 1, 0, 0},
               {"strict_draftn", 1, 0, 0},
               {"enable_okc", 1, 0, 0},
               {"verbose", 0, 0, 0},
               {"setap_cfg",1,0,0},
#ifdef CSR_ONLY_DEBUG
               /* Softmac native driver */
               {"add_template", 1, 0, 0},
               {"del_template", 1, 0, 0},
               {"set_filter", 1, 0, 0},
#endif /* CSR_ONLY_DEBUG */
               {0, 0, 0, 0} };

    static char * const filter_options[] = {"dhcp", "arp", "nbns", "nbds",
                                           "cups", "none", "all", NULL };

    static char * const uapsd_options[] = {"vo", "vi", "bk", "be", "none",
                                          "all", "default", NULL };

    static char * const maxsp_options[] = {"2", "4", "6", "all", NULL };

#ifdef CSR_ONLY_DEBUG
    /* Softmac native driver */
    static char * const filter_action_options[] = {"ether", "udi", "all", "none", NULL };
#endif /* CSR_ONLY_DEBUG */
    if (argc > 1) {
        if (!strcasecmp(argv[1], "--help")) {
            print_usage();
            exit(1);
        }
    }

    start = 0;
    show_set = 0;
    wifi_filter_set = 0;
    wifi_filter = UNIFI_CFG_FILTER_NONE;
    uapsd_mask_set = uapsd_mask = 0;
    maxsp_set = maxsp_length = wmm_qos_info = 0;
    wmm_addts_tid_set = addts_ie_length = wmm_addts_ie_set = wmm_addts_tid = 0;
    wmm_delts_tid_set = wmm_delts_tid = 0;
    file_name[0] =0;
    phymode[0] = 0;

    while ((optc = getopt_long(argc, argv, "", long_options, &option_index)) != -1) {
        start = 1;
        if (optc == 0) {

            if (!strcasecmp(long_options[option_index].name, "verbose")) {
                verbose++;
                continue;
            }

            if (!strcasecmp(long_options[option_index].name, "dev")) {
                size_t sz = strlen(optarg);
                if (sz >= (sizeof(device) - 1)) {
                    printf("dev arg %s is too long (%u)\n", optarg, sz);
                    exit(1);
                }
                memcpy(device, optarg, sz);
                device[sz] = 0;
                continue;
            }

            if (!strcasecmp(long_options[option_index].name, "wifion")) {
                wifi_power = UNIFI_CFG_POWER_ON;
                continue;
            }

            if (!strcasecmp(long_options[option_index].name, "wifioff")) {
                wifi_power = UNIFI_CFG_POWER_OFF;
                continue;
            }

            if (!strcasecmp(long_options[option_index].name, "show")) {
                show_set = 1;
                continue;
            }

            if (!strcasecmp(long_options[option_index].name, "batteries")) {

                if (optarg == NULL) {
                    printf("option batteries requires an argument\n");
                    exit(1);
                }

                if (!strcasecmp(optarg, "yes")) {
                    wifi_powersupply = UNIFI_CFG_POWERSUPPLY_BATTERIES;
                } else if (!strcasecmp(optarg, "no")) {
                    wifi_powersupply = UNIFI_CFG_POWERSUPPLY_MAINS;
                } else {
                    printf("batteries invalid option: %s\n", optarg);
                    start = 0;
                }
                continue;
            }
            if (!strcasecmp(long_options[option_index].name, "strict_draftn")) {

                if (optarg == NULL) {
                    printf("option strict_draftn requires an argument\n");
                    exit(1);
                }

                if (!strcasecmp(optarg, "yes")) {
                    strict_draftn = TRUE;
                    strict_draftn_set = TRUE;
                } else if (!strcasecmp(optarg, "no")) {
                    strict_draftn = FALSE;
                    strict_draftn_set = TRUE;
                } else {
                    printf("strict_draftn invalid option: %s\n", optarg);
                    start = 0;
                }
                continue;
            }

            if (!strcasecmp(long_options[option_index].name, "enable_okc")) {

                if (optarg == NULL) {
                    printf("option enable_okc requires an argument\n");
                    exit(1);
                }

                if (!strcasecmp(optarg, "yes")) {
                    enable_okc = TRUE;
                    enable_okc_set = TRUE;
                } else if (!strcasecmp(optarg, "no")) {
                    enable_okc = FALSE;
                    enable_okc_set = TRUE;
                } else {
                    printf("enable_okc invalid option: %s\n", optarg);
                    start = 0;
                }
                continue;
            }
            if (!strcasecmp(long_options[option_index].name, "powersave")) {

                if (optarg == NULL) {
                    printf("option powersave requires an argument\n");
                    exit(1);
                }

                if (!strcasecmp(optarg, "none")) {
                    wifi_powersave = UNIFI_CFG_POWERSAVE_NONE;
                } else if (!strcasecmp(optarg, "fast")) {
                    wifi_powersave = UNIFI_CFG_POWERSAVE_FAST;
                } else if (!strcasecmp(optarg, "full")) {
                    wifi_powersave = UNIFI_CFG_POWERSAVE_FULL;
                } else if (!strcasecmp(optarg, "auto")) {
                    wifi_powersave = UNIFI_CFG_POWERSAVE_AUTO;
                } else {
                    printf("powersave invalid option: %s\n", optarg);
                    start = 0;
                }
                continue;
            }

            if (!strcasecmp(long_options[option_index].name, "filter")) {

                while ((index = getsubopt(&optarg, filter_options, &value)) != -1) {
                    switch(index) {
                      case 0:
                        wifi_filter |= UNIFI_CFG_FILTER_DHCP;
                        wifi_filter_set = 1;
                        break;
                      case 1:
                        wifi_filter |= UNIFI_CFG_FILTER_ARP;
                        wifi_filter_set = 1;
                        break;
                      case 2:
                        wifi_filter |= UNIFI_CFG_FILTER_NBNS;
                        wifi_filter_set = 1;
                        break;
                      case 3:
                        wifi_filter |= UNIFI_CFG_FILTER_NBDS;
                        wifi_filter_set = 1;
                        break;
                      case 4:
                        wifi_filter |= UNIFI_CFG_FILTER_CUPS;
                        wifi_filter_set = 1;
                        break;
                      case 5:
                        wifi_filter = UNIFI_CFG_FILTER_NONE;
                        wifi_filter_set = 1;
                        break;
                      case 6:
                        wifi_filter = UNIFI_CFG_FILTER_ALL;
                        wifi_filter_set = 1;
                        break;
                      default:
                        printf("unrecognised filter arg: %s\n", filter_options[index]);
                        break;
                    }
                }

                if (!wifi_filter_set)
                {
                    printf("option --filter requires an argument\n");
                    exit(1);
                }
                continue;
            }

            if (!strcasecmp(long_options[option_index].name, "opt_in")) {
                wifi_filter_opt_in_flag = 1;
                continue;
            }

            if (!strcasecmp(long_options[option_index].name, "core_dump")) {
                fw_core_dump_flag = 1;
                continue;
            }

            if (!strcasecmp(long_options[option_index].name, "uapsd")) {

                while ((index = getsubopt(&optarg, uapsd_options, &value)) != -1) {
                    switch(index) {
                      case 0:       /* VO */
                      case 1:       /* VI */
                      case 2:       /* BK */
                      case 3:       /* BE */
                        uapsd_mask |= (1 << index);
                        uapsd_mask_set = 1;
                        break;
                      case 4:
                        uapsd_mask = 0x00;
                        uapsd_mask_set = 1;
                        break;
                      case 5:
                        uapsd_mask = 0x0F;      /* All 4 bits set to 1 */
                        uapsd_mask_set = 1;
                        break;
                      case 6:
                        uapsd_mask = 0xFF;      /* SME will handle this */
                        uapsd_mask_set = 1;
                        break;
                      default:
                        printf("unrecognised filter arg: %s\n", uapsd_options[index]);
                        break;
                    }
                }

                if (!uapsd_mask_set)
                {
                    printf("option --uapsd requires an argument\n");
                    exit(1);
                }
                continue;
            }

            if (!strcasecmp(long_options[option_index].name, "maxsp")) {

                while ((index = getsubopt(&optarg, maxsp_options, &value)) != -1) {
                    switch(index) {
                      case 0:
                        maxsp_length = 0x01;
                        maxsp_set = 1;
                        break;
                      case 1:
                        maxsp_length = 0x02;
                        maxsp_set = 1;
                        break;
                      case 2:
                        maxsp_length = 0x03;
                        maxsp_set = 1;
                        break;
                      case 3:
                        maxsp_length = 0x00;
                        maxsp_set = 1;
                        break;
                      default:
                        printf("unrecognised filter arg: %s\n", maxsp_options[index]);
                        break;
                    }
                }

                if (!maxsp_set)
                {
                    printf("option --maxsp requires an argument\n");
                    exit(1);
                }
                continue;
            }

            if (!strcasecmp(long_options[option_index].name, "addts_ie")) {
                addts_ie_length = read_ie(optarg, addts_ie);
                if (addts_ie_length != UNIFI_CFG_MAX_ADDTS_IE_LENGTH) {
                    printf("option --addts_ie requires a valid %d bytes hex array\n",
                           UNIFI_CFG_MAX_ADDTS_IE_LENGTH);
                    exit(1);
                }
                wmm_addts_ie_set = 1;
                continue;
            }

            if (!strcasecmp(long_options[option_index].name, "addts_tid")) {
                if (!read_byte(optarg, &uchar_value)) {
                    printf("option --addts_tid requires a valid hex number\n");
                    exit(1);
                }
                wmm_addts_tid = uchar_value;
                wmm_addts_tid_set = 1;
                continue;
            }

            if (!strcasecmp(long_options[option_index].name, "delts_tid")) {
                if (!read_byte(optarg, &uchar_value)) {
                    printf("option --addts_tid requires a valid hex number\n");
                    exit(1);
                }
                wmm_delts_tid = uchar_value;
                wmm_delts_tid_set = 1;
                continue;
            }
            if (!strcasecmp(long_options[option_index].name, "setap_cfg")) {
                size_t sz = strlen(optarg);
                if (sz >= (UNIFI_CFG_MAX_FILE_LENGTH-1)) {
                    printf("setap_cfg arg %s is too long (%u)\n", optarg, sz);
                    exit(1);
                }
                memcpy(file_name, optarg, sz);
                file_name[sz] = 0;
                ap_cfg = TRUE;
                continue;
            }
#ifdef CSR_ONLY_DEBUG
            /* Softmac native driver */
            if (strcasecmp(long_options[option_index].name, "add_template") == 0
             || strcasecmp(long_options[option_index].name, "del_template") == 0) {
                /* Only one of add_template or del_template may be specified */
                if (add_template || del_template)
                {
                    printf("options --add_template and --del_template cannot be duplicated or mixed\n");
                    exit(1);
                }
                /* MAC address specified? */
                if (strchr(optarg, ':'))
                {
                    /* Decode MAC address... */
                    template_mac = malloc(6);
                    if (read_byte(optarg+0, &template_mac[0]) != 2 || optarg[2] != ':'
                     || read_byte(optarg+3, &template_mac[1]) != 2 || optarg[5] != ':'
                     || read_byte(optarg+6, &template_mac[2]) != 2 || optarg[8] != ':'
                     || read_byte(optarg+9, &template_mac[3]) != 2 || optarg[11] != ':'
                     || read_byte(optarg+12, &template_mac[4]) != 2 || optarg[14] != ':'
                     || read_byte(optarg+15, &template_mac[5]) != 2 || optarg[17] != '\0')
                    {
                        printf("option %s can't grok '%s' as a MAC address\n", long_options[option_index].name, optarg);
                        exit(1);
                    }
                    if (optind == argc) {
                        printf("option %s missing PRIORITY\n", long_options[option_index].name);
                        exit(1);
                    }
                    optarg=argv[optind++];
                }
                else {
                    /* Wildcard */
                    template_mac = 0;
                }
                /* Priority */
                template_priority = strtol(optarg,0,16);
                if (strcasecmp(long_options[option_index].name, "add_template") == 0) {
                    /* Template */
                    if (optind == argc) {
                        printf("option %s missing TEMPLATE\n", long_options[option_index].name);
                        exit(1);
                    }
                    optarg=argv[optind++];
                    if (strlen(optarg) & 1)
                    {
                        printf("option %s TEMPLATE has odd length\n", long_options[option_index].name);
                        exit(1);
                    }
                    if (strlen(optarg) >= 0x200)
                    {
                        printf("option %s TEMPLATE is too big\n", long_options[option_index].name);
                        exit(1);
                    }
                    template_len = strlen(optarg)/2;
                    template_data = malloc(template_len);
                    for (uchar_value=0; uchar_value<template_len; uchar_value++) {
                        if (read_byte(optarg+(2*uchar_value), template_data+uchar_value) != 2) {
                            printf("option %s can't grok '%s'[%d] as a TEMPLATE\n", long_options[option_index].name, optarg, 2*uchar_value);
                            exit(1);
                        }
                    }

                    add_template = 1;
                }
                else {
                    del_template = 1;
                }
                continue;
            }
            if (strcasecmp(long_options[option_index].name, "set_filter") == 0) {
                filter_precedence = strtol(optarg,0,10);
                if (optind == argc) {
                    printf("option %s missing ACTION\n", long_options[option_index].name);
                    exit(1);
                }
                optarg=argv[optind++];
                while ((index = getsubopt(&optarg, filter_action_options, &value)) != -1) {
                    switch(index) {
                      case 0:
                        filter_action |= 0x01;
                        break;
                      case 1:
                        filter_action |= 0x02;
                        break;
                      case 2:   /* all */
                        filter_action = 0xFF;
                        break;
                      case 3:   /* none */
                        filter_action = 0x00;
                        break;
                      default:
                        printf("unrecognised action arg: %s\n", maxsp_options[index]);
                        break;
                    }
                }
                if (optind == argc) {
                    printf("option %s missing TEMPLATE\n", long_options[option_index].name);
                    exit(1);
                }
                optarg=argv[optind++];
                if (strlen(optarg) & 1)
                {
                    printf("option %s TEMPLATE has odd length\n", long_options[option_index].name);
                    exit(1);
                }
                if (strlen(optarg) >= 0x200)
                {
                    printf("option %s TEMPLATE is too big\n", long_options[option_index].name);
                    exit(1);
                }
                filter_len = strlen(optarg)/2;
                filter_content = malloc(filter_len);
                filter_mask = malloc(filter_len);
                for (uchar_value=0; uchar_value<filter_len; uchar_value++) {
                    filter_mask[uchar_value] = 0xFF;
                    if (optarg[2*uchar_value] == '_')
                    {
                        filter_mask[uchar_value] &= 0x0F;
                        optarg[2*uchar_value] = '0';
                    }
                    if (optarg[2*uchar_value+1] == '_')
                    {
                        filter_mask[uchar_value] &= 0xF0;
                        optarg[2*uchar_value+1] = '0';
                    }
                    if (read_byte(optarg+(2*uchar_value), filter_content+uchar_value) != 2) {
                        printf("option %s can't grok '%s'[%d] as a TEMPLATE\n", long_options[option_index].name, optarg, 2*uchar_value);
                        exit(1);
                    }
                }

                set_filter = 1;
                continue;
            }
#endif /* CSR_ONLY_DEBUG */

            printf("Unrecognised option: %s\n", long_options[option_index].name);
            start = 0;
            break;

        } else {
            start = 0;
        }
    }

    if (optind < argc)
    {
        printf("Unrecognised option(s) ignored: ");
        while (optind < argc) {
            printf ("%s ", argv[optind++]);
        }
        printf ("\n");
    }

    if (start == 0) {
        print_usage();
        exit(1);
    }

    if (verbose) {
        fprintf(stderr, " Using:\n");
        fprintf(stderr, " DEVICE: %s\n", device);
        fprintf(stderr, " Show: %s\n", (show_set) ? "YES" : "NO");
        fprintf(stderr, " WiFi ON/OFF: %d\n", wifi_power);
        fprintf(stderr, " Power: %d\n", wifi_powersave);
        fprintf(stderr, " Filter: 0x%x\n", wifi_filter);
        fprintf(stderr, " U-APSD Mask: 0x%x\n", uapsd_mask);
        if(ap_cfg) {
            fprintf(stderr, " AP Config File Name: %s\n", file_name);
        }
        fprintf(stderr, "\n");
    }

    fd = unifi_cfg_init(device);
    if (fd < 0) {
        perror("Failed to open device");
        exit(errno);
    }

    if (strict_draftn_set) {
        r = unifi_cfg_set_strict_draft_n(fd, strict_draftn);
        if (r < 0) {
            printf("UNIFI_CFG_STRICT_DRAFT_N: failed\n");
            goto EXIT;
        }
    }
    if (enable_okc_set) {
        r = unifi_cfg_set_enable_okc(fd, enable_okc);
        if (r < 0) {
            printf("UNIFI_CFG_ENABLE_OKC: failed\n");
            goto EXIT;
        }
    }
    if (wifi_power != UNIFI_CFG_POWER_UNSPECIFIED) {
        r = unifi_cfg_set_power(fd, wifi_power);
        if (r < 0) {
            printf("UNIFI_CFG_POWER: failed\n");
            goto EXIT;
        }
    }

    if (wifi_powersave != UNIFI_CFG_POWERSAVE_UNSPECIFIED) {
        r = unifi_cfg_set_power_save(fd, wifi_powersave);
        if (r < 0) {
            printf("UNIFI_CFG_POWERSAVE: failed\n");
            goto EXIT;
        }
    }

    if (wifi_powersupply != UNIFI_CFG_POWERSUPPLY_UNSPECIFIED) {
        r = unifi_cfg_set_power_supply(fd, wifi_powersupply);
        if (r < 0) {
            printf("UNIFI_CFG_POWERSUPPLY: failed\n");
            goto EXIT;
        }
    }

    if (wifi_filter_set) {
        r = unifi_cfg_set_filters(fd, wifi_filter, wifi_filter_opt_in_flag);
        if (r < 0) {
            printf("UNIFI_CFG_FILTER: failed\n");
            goto EXIT;
        }
    }

    if (fw_core_dump_flag) {
        r = unifi_cfg_trigger_core_dump(fd);
        if (r < 0) {
            printf("UNIFI_CFG core dump failed\n");
            goto EXIT;
        }
    }

    if (maxsp_set || uapsd_mask_set) {
        /* Max SP Length can only be used with one or more U-APSD ACs */
        if (!uapsd_mask && maxsp_set) {
            printf("UNIFI_CFG_WMM_QOSINFO: Max SP Length can only be used with one or more U-APSD ACs\n");
            goto EXIT;
        }

        wmm_qos_info = (maxsp_length << 5) | uapsd_mask;
        r = unifi_cfg_set_wmm_qos_info(fd, wmm_qos_info);
        if (r < 0) {
            printf("UNIFI_CFG_WMM_QOSINFO: failed\n");
            goto EXIT;
        }
    }

    if (wmm_delts_tid_set) {
        r = unifi_cfg_wmm_delts(fd, wmm_delts_tid);
        if (r < 0) {
            printf("UNIFI_CFG_WMM_DELTS: failed\n");
            goto EXIT;
        }
    }

    if (wmm_addts_tid_set && wmm_addts_ie_set) {
        r = unifi_cfg_wmm_addts(fd, wmm_addts_tid, addts_ie, addts_ie_length);
        if (r < 0) {
            printf("UNIFI_CFG_WMM_ADDTS: failed\n");
            goto EXIT;
        }
    }
    if(ap_cfg) {
        r = unifi_cfg_set_ap_config(fd,file_name);
        if (r < 0) {
            printf("UNIFI_CFG_SET_AP_CONFIG: failed possibily the driver does not support AP configuration\n");
            goto EXIT;
        }
    }
#ifdef CSR_ONLY_DEBUG
    /* Softmac native driver */
    if (add_template) {
        r = unifi_cfg_add_softmac_template(fd, template_mac, template_priority, template_len, template_data);
        if (r < 0) {
            printf("UNIFI_SOFTMAC_MA_PACKET_TEMPLATE (add): failed\n");
            goto EXIT;
        }
    }
    if (del_template) {
        r = unifi_cfg_del_softmac_template(fd, template_mac, template_priority);
        if (r < 0) {
            printf("UNIFI_SOFTMAC_MA_PACKET_TEMPLATE (del): failed\n");
            goto EXIT;
        }
    }
    if (set_filter) {
        r = unifi_cfg_set_softmac_filter(fd, filter_precedence, filter_action, filter_len, filter_content, filter_mask);
        if (r < 0) {
            printf("UNIFI_SOFTMAC_MA_PACKET_FILTER (set): failed\n");
            goto EXIT;
        }
    }
#endif /* CSR_ONLY_DEBUG */

    if (show_set) {
        union unifi_cfg_get_info_arg param;

        /* Interface name */
        param.drv_inst[0] = '\0';
        r = unifi_cfg_get_info(fd, UNIFI_CFG_GET_INSTANCE, (unsigned char *)param.drv_inst);
        if (r < 0) {
            printf("UNIFI_CFG_GET_INSTANCE: failed\n");
        }

        printf("Current Driver Info:\n");
        printf("    Interface Name = %s\n", param.drv_inst);

        /* Co-existence */
        r = unifi_cfg_get_info(fd, UNIFI_CFG_GET_COEX, (unsigned char *)&param.coex_info);
        if (r < 0) {
            printf("UNIFI_CFG_GET_COEX: failed\n");
            goto EXIT;
        }

        printf("Current Coex Info:\n");
        printf("    Traffic Type = %s\n",
               trace_traffic_type(param.coex_info.currentTrafficType));
        if (param.coex_info.currentTrafficType == CSR_WIFI_SME_TRAFFIC_TYPE_PERIODIC) {
            printf("    Traffic Period = %d\n", param.coex_info.currentPeriodMs);
        } else if (param.coex_info.currentCoexLatencyMs) {
            printf("    Traffic Latency = %d\n", param.coex_info.currentCoexLatencyMs);
        }

        printf("    Power Save Mode = %s\n",
               trace_power_save_type(param.coex_info.currentPowerSave));
        printf("    Has Traffic Data = %s\n",
               param.coex_info.hasTrafficData ? "Yes" : "No");
        printf("    Has Bluetooth device = %s\n",
               param.coex_info.hasBtDevice ? "Yes" : "No");
        printf("    Coex Scheme = %s\n",
               trace_coex_scheme_type(param.coex_info.currentCoexScheme));

        /* Power save mode */
        r = unifi_cfg_get_info(fd, UNIFI_CFG_GET_POWER_MODE, (unsigned char *)&param.power_save_info);
        if (r < 0) {
            printf("UNIFI_CFG_GET_POWER_MODE: failed\n");
            goto EXIT;
        }

        printf("Configured Power Save Mode = %s\n",
               trace_power_save_type(param.power_save_info));

        /* Power supply mode */
        r = unifi_cfg_get_info(fd, UNIFI_CFG_GET_POWER_SUPPLY, (unsigned char *)&param.host_power_mode);
        if (r < 0) {
            printf("UNIFI_CFG_GET_POWER_SUPPLY: failed\n");
            goto EXIT;
        }
        printf("%s\n",
               trace_power_supply_type(param.host_power_mode));

        r = unifi_cfg_get_info(fd, UNIFI_CFG_GET_AP_CONFIG, (unsigned char *)&param.ap_cfg);
        if(r == -EPERM) {
            /* This is received when AP operation is tried and driver does not
               support it */
            r = 0;
            goto EXIT;
        }
        if (r < 0) {
            printf("UNIFI_CFG_GET_AP_CONFIG: failed \n");
            goto EXIT;
        }
        printf("\nAP Configuration\n");
        printf("    Beacon Interval = %d TU\n",param.ap_cfg.beaconInterval);
        printf("    DTIM Period = %d BI\n",param.ap_cfg.dtimPeriod);
        printf("    WMM %s\n",(param.ap_cfg.wmmEnabled)?"Enabled":"Disabled");
        if(param.ap_cfg.phySupportedBitmap & (1<<0))
            strcat(phymode,"a");
        if(param.ap_cfg.phySupportedBitmap & (1<<1))
            strcat(phymode,"b");
        if(param.ap_cfg.phySupportedBitmap & (1<<2))
            strcat(phymode,"g");
        if(param.ap_cfg.phySupportedBitmap & (1<<3))
            strcat(phymode,"n");
        printf("    PhyMode = %s\n",phymode);
    }


EXIT:
    /* If the command failed report the error. Generally this will be the
     * driver error code returned via the ioctl.
     */
    if (r < 0) {
        printf("status %d, errno %d\n", r, errno);
    }

    /* Kill communication with driver. */
    unifi_cfg_deinit(fd);

    return r;
} /* main() */

