/*
 * ***************************************************************************
 *
 *  FILE: unifi_config_lib.c
 *
 *      Get/Set configuration to the UniFi driver library.
 *
 * Copyright (C) 2008 by Cambridge Silicon Radio Ltd.
 *
 * Refer to LICENSE.txt included with this source code for details on
 * the license terms.
 *
 * ***************************************************************************
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <getopt.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#include <stddef.h>

#include "csr_types.h"
#include "csr_wifi_sme_prim.h"
#include "unifiio.h"
#include "unifi_config_lib.h"

#if 0
static void
dump(unsigned char* in_buffer, unsigned int count)
{
    unsigned int i;
    for (i=0; i<count; i++) {
        printf("%02X ", in_buffer[i]);
        if ((i % 16) == 15) {
            printf("\n");
        }
    }
    printf("\n");
}
#endif

static void TcpIp4Tclas(tclas_t* tclas)
{
    memset(tclas, 0, sizeof(tclas_t));
    tclas->element_id = 14;
    tclas->length = sizeof(tcpip_clsfr_t) + 1;
    tclas->user_priority = 0;
    tclas->tcp_ip_cls_fr.cls_fr_type = 1;
    tclas->tcp_ip_cls_fr.version = 4;
}

static CsrUint32
FilterCups(tclas_t* tclas)
{
    TcpIp4Tclas(tclas);
    ((CsrUint8*)(&tclas->tcp_ip_cls_fr.source_port))[0] = 0x02;
    ((CsrUint8*)(&tclas->tcp_ip_cls_fr.source_port))[1] = 0x77;
    ((CsrUint8*)(&tclas->tcp_ip_cls_fr.dest_port))[0] = 0x02;
    ((CsrUint8*)(&tclas->tcp_ip_cls_fr.dest_port))[1] = 0x77;
    tclas->tcp_ip_cls_fr.protocol = 0x11;
    tclas->tcp_ip_cls_fr.cls_fr_mask = 0x58; //bits: 3,4,6
    return sizeof(tclas_t);
}

static CsrUint32
FilterNbns(tclas_t* tclas)
{
    TcpIp4Tclas(tclas);
    ((CsrUint8*)(&tclas->tcp_ip_cls_fr.source_port))[0] = 0x00;
    ((CsrUint8*)(&tclas->tcp_ip_cls_fr.source_port))[1] = 0x89;
    ((CsrUint8*)(&tclas->tcp_ip_cls_fr.dest_port))[0] = 0x00;
    ((CsrUint8*)(&tclas->tcp_ip_cls_fr.dest_port))[1] = 0x89;
    tclas->tcp_ip_cls_fr.protocol = 0x11;
    tclas->tcp_ip_cls_fr.cls_fr_mask = 0x58; //bits: 3,4,6

    return sizeof(tclas_t);
}

static CsrUint32
FilterNbds(tclas_t* tclas)
{
    TcpIp4Tclas(tclas);
    ((CsrUint8*)(&tclas->tcp_ip_cls_fr.source_port))[0] = 0x00;
    ((CsrUint8*)(&tclas->tcp_ip_cls_fr.source_port))[1] = 0x8A;
    ((CsrUint8*)(&tclas->tcp_ip_cls_fr.dest_port))[0] = 0x00;
    ((CsrUint8*)(&tclas->tcp_ip_cls_fr.dest_port))[1] = 0x8A;
    tclas->tcp_ip_cls_fr.protocol = 0x11;
    tclas->tcp_ip_cls_fr.cls_fr_mask = 0x58; //bits: 3,4,6
    return sizeof(tclas_t);
}

static int
_add_filter(unsigned char* in_buffer, unsigned int filter, int opt_in_flag)
{
    uf_cfg_bcast_packet_filter_t *bcast_packet_filter;
    unsigned char *tclas;
    unsigned int bcast_packet_filter_size;
    unsigned int tclas_ies_length;
    CsrUint32 added_tclas_length;

    bcast_packet_filter = (uf_cfg_bcast_packet_filter_t*)in_buffer;

    bcast_packet_filter->filter_mode = (opt_in_flag? 3: 0); /* Unit data filter mode from HIP spec */
    tclas = bcast_packet_filter->tclas_ies;
    bcast_packet_filter_size = sizeof(uf_cfg_bcast_packet_filter_t);
    tclas_ies_length = 0;

    bcast_packet_filter->arp_filter = filter & UNIFI_CFG_FILTER_ARP;
    bcast_packet_filter->dhcp_filter = filter & UNIFI_CFG_FILTER_DHCP;

    if(filter & UNIFI_CFG_FILTER_CUPS) {
        added_tclas_length = FilterCups((tclas_t*)(tclas + tclas_ies_length));
        tclas_ies_length += added_tclas_length;
        bcast_packet_filter_size += added_tclas_length;
    }

    if(filter & UNIFI_CFG_FILTER_NBNS) {
        added_tclas_length = FilterNbns((tclas_t*)(tclas + tclas_ies_length));
        tclas_ies_length += added_tclas_length;
        bcast_packet_filter_size += added_tclas_length;
    }

    if(filter & UNIFI_CFG_FILTER_NBDS) {
        added_tclas_length = FilterNbds((tclas_t*)(tclas + tclas_ies_length));
        tclas_ies_length += added_tclas_length;
        bcast_packet_filter_size += added_tclas_length;
    }

    bcast_packet_filter->tclas_ies_length = tclas_ies_length;

    /*
    dump(bcast_packet_filter, sizeof(uf_cfg_bcast_packet_filter_t));
    dump(tclas, bcast_packet_filter->tclas_ies_length);
    */

    return (tclas_ies_length) ? (bcast_packet_filter_size - 1) : bcast_packet_filter_size;
}


int unifi_cfg_init(char *device)
{
    int fd;

    /* Start communication with driver. */
    fd = open(device, O_RDWR);
    if (fd < 0) {
        printf("Failed to open device\n");
    }
    return fd;
}

int unifi_cfg_deinit(int fd)
{
    /* Kill communication with driver. */
    close(fd);

    return 0;
}

int unifi_cfg_set_filters(int fd, unsigned int filter_type, int opt_in_flag)
{
    int r, filter_size;
    unsigned char buffer[128];
    unifi_cfg_command_t *pcmd = (unifi_cfg_command_t *)buffer;

    *pcmd = UNIFI_CFG_FILTER;

    filter_size = _add_filter(buffer + sizeof(unsigned int) + sizeof(unifi_cfg_command_t), filter_type, opt_in_flag);
    *((unsigned int*)(buffer + sizeof(unifi_cfg_command_t))) = filter_size;

    r = ioctl(fd, UNIFI_CFG, buffer);
    if (r < 0) {
        printf("UNIFI_CFG: failed\n");
        return r;
    }

    return 0;
}


int unifi_cfg_trigger_core_dump(int fd)
{
    int r;
    unifi_cfg_command_t cmd[1];

    cmd[0] = UNIFI_CFG_CORE_DUMP;

    r = ioctl(fd, UNIFI_CFG, (unsigned char*)cmd);
    if (r < 0) {
        printf("unifi_cfg_trigger_core_dump: ioctl failed\n");
        return r;
    }

    return 0;
}


int unifi_cfg_set_wmm_qos_info(int fd, CsrUint8 wmm_qos_info)
{
    int r;
    unsigned char buffer[8];
    unifi_cfg_command_t *pcmd = (unifi_cfg_command_t *)buffer;

    *pcmd = UNIFI_CFG_WMM_QOSINFO;
    *((CsrUint8*)(buffer + sizeof(unifi_cfg_command_t))) = wmm_qos_info;

    r = ioctl(fd, UNIFI_CFG, buffer);
    if (r < 0) {
        printf("unifi_cfg_set_wmm_qos_info: failed\n");
        return r;
    }

    return 0;
}


int unifi_cfg_set_power_save(int fd, unifi_cfg_powersave_t power_mode)
{
    int r;
    unsigned char buffer[16];
    unifi_cfg_command_t *pcmd = (unifi_cfg_command_t *)buffer;

    *pcmd = UNIFI_CFG_POWERSAVE;
    *((unifi_cfg_powersave_t*)(buffer + sizeof(unifi_cfg_command_t))) = power_mode;

    r = ioctl(fd, UNIFI_CFG, buffer);
    if (r < 0) {
        printf("unifi_cfg_set_power_save: failed\n");
        return r;
    }

    return 0;
}

int unifi_cfg_set_power(int fd, unifi_cfg_power_t power)
{
    int r;
    unsigned char buffer[16];
    unifi_cfg_command_t *pcmd = (unifi_cfg_command_t *)buffer;

    *pcmd = UNIFI_CFG_POWER;
    *((unifi_cfg_power_t*)(buffer + sizeof(unifi_cfg_command_t))) = power;

    r = ioctl(fd, UNIFI_CFG, buffer);
    if (r < 0) {
        printf("unifi_cfg_set_power: failed\n");
        return r;
    }

    return 0;
}

int unifi_cfg_set_power_supply(int fd, unifi_cfg_powersupply_t power_supply)
{
    int r;
    unsigned char buffer[16];
    unifi_cfg_command_t *pcmd = (unifi_cfg_command_t *)buffer;

    *pcmd = UNIFI_CFG_POWERSUPPLY;
    *((unifi_cfg_powersupply_t*)(buffer + sizeof(unifi_cfg_command_t))) = power_supply;

    r = ioctl(fd, UNIFI_CFG, buffer);
    if (r < 0) {
        printf("unifi_cfg_set_power_supply: failed\n");
        return r;
    }

    return 0;
}


int unifi_cfg_wmm_addts(int fd, CsrUint32 addts_tid,
                        CsrUint8 *addts_ie, CsrUint8 addts_ie_length)
{
    int r;
    static unsigned char buffer[UNIFI_CFG_MAX_ADDTS_IE_LENGTH + 16];
    unifi_cfg_command_t *pcmd = (unifi_cfg_command_t *)buffer;
    CsrUint8 *buf_pos;

    if (addts_ie_length > UNIFI_CFG_MAX_ADDTS_IE_LENGTH) {
        printf("unifi_cfg_wmm_addts: IE exceeds UNIFI_CFG_MAX_ADDTS_IE_LENGTH\n");
        return -EINVAL;
    }

    buf_pos = buffer;
    *pcmd = UNIFI_CFG_WMM_ADDTS;

    buf_pos += sizeof(unifi_cfg_command_t);
    *((CsrUint32*)buf_pos) = addts_tid;

    buf_pos += sizeof(CsrUint32);
    *((CsrUint8*)buf_pos) = addts_ie_length;

    buf_pos += sizeof(CsrUint8);
    memcpy(buf_pos, addts_ie, addts_ie_length);

    r = ioctl(fd, UNIFI_CFG, buffer);
    if (r < 0) {
        printf("unifi_cfg_wmm_addts: failed\n");
        return r;
    }

    return 0;
}


int unifi_cfg_wmm_delts(int fd, CsrUint32 delts_tid)
{
    int r;
    static unsigned char buffer[8];
    unifi_cfg_command_t *pcmd = (unifi_cfg_command_t *)buffer;
    CsrUint8 *buf_pos;

    buf_pos = buffer;
    *pcmd = UNIFI_CFG_WMM_DELTS;

    buf_pos += sizeof(unifi_cfg_command_t);
    *((CsrUint32*)buf_pos) = delts_tid;

    r = ioctl(fd, UNIFI_CFG, buffer);
    if (r < 0) {
        printf("unifi_cfg_wmm_delts: failed\n");
        return r;
    }

    return 0;
}

static void unifi_cfg_set_ap_default_config(uf_cfg_ap_config_t * ap_cfg)
{
    ap_cfg->phySupportedBitmap = 6;
    ap_cfg->channel = 6;
    ap_cfg->beaconInterval = 100;
    ap_cfg->dtimPeriod = 3;
    ap_cfg->wmmEnabled = 1;
    ap_cfg->shortSlotTimeEnabled = 1;
    ap_cfg->groupkeyTimeout = 0;
    ap_cfg->strictGtkRekeyEnabled = 0;
    ap_cfg->gmkTimeout = 0;
    ap_cfg->responseTimeout =200;
    ap_cfg->retransLimit = 3;
    ap_cfg->rxStbc = 1;
    ap_cfg->rifsModeAllowed = 0;
    ap_cfg->dualCtsProtection = 0;
    ap_cfg->ctsProtectionType = 0;
    ap_cfg->maxListenInterval = 255;
}

int unifi_cfg_set_ap_config(int fd,const char* fileName)
{

    FILE *fp;
    char buf[256], *pos;
    int line = 0;
    int r;
    unsigned char buffer[128];
    uf_cfg_ap_config_t * ap_cfg = (uf_cfg_ap_config_t *)(buffer + sizeof(unsigned int) + sizeof(unifi_cfg_command_t));
    unifi_cfg_command_t *pcmd = (unifi_cfg_command_t *)buffer;

    *pcmd = UNIFI_CFG_SET_AP_CONFIG;

    *((unsigned int*)(buffer + sizeof(unifi_cfg_command_t))) = sizeof(uf_cfg_ap_config_t);

    fp = fopen(fileName, "r");
    if (fp == NULL) {
        printf("Could not open configuration file '%s' "
			   "for reading.", fileName);
        return -1;
    }
    unifi_cfg_set_ap_default_config(ap_cfg);
    while (fgets(buf, sizeof(buf), fp)) {
        line++;

        if (buf[0] == '#')
            continue;
        pos = buf;
        while (*pos != '\0') {
            if (*pos == '\n') {
                *pos = '\0';
                break;
            }
            pos++;
        }
        if (buf[0] == '\0')
            continue;

        pos = strchr(buf, '=');
        if (pos == NULL) {
            printf("Line %d: invalid line '%s'",line, buf);
            continue;
        }
        *pos = '\0';
        pos++;
        if (strcmp(buf, "phySupportedBitmap") == 0) {
           ap_cfg->phySupportedBitmap = atoi(pos);
        } else if (strcmp(buf, "channel") == 0) {
           ap_cfg->channel = atoi(pos);
        } else if (strcmp(buf, "beaconInterval") == 0) {
           ap_cfg->beaconInterval = atoi(pos);
        } else if (strcmp(buf, "dtimPeriod") == 0) {
           ap_cfg->dtimPeriod = atoi(pos);
        } else if (strcmp(buf, "wmmEnabled") == 0) {
           ap_cfg->wmmEnabled = atoi(pos);
        } else if (strcmp(buf, "shortSlotTimeEnabled") == 0) {
           ap_cfg->shortSlotTimeEnabled = atoi(pos);
        } else if (strcmp(buf, "ctsProtectionType") == 0) {
           ap_cfg->ctsProtectionType = atoi(pos);
        } else if (strcmp(buf, "groupkeyTimeout") == 0) {
           ap_cfg->groupkeyTimeout = atoi(pos);
        } else if (strcmp(buf, "strictGtkRekeyEnabled") == 0) {
           ap_cfg->strictGtkRekeyEnabled = atoi(pos);
        } else if (strcmp(buf, "gmkTimeout") == 0) {
           ap_cfg->gmkTimeout = atoi(pos);
        } else if (strcmp(buf, "responseTimeout") == 0) {
           ap_cfg->responseTimeout = atoi(pos);
        } else if (strcmp(buf, "retransLimit") == 0) {
           ap_cfg->retransLimit = atoi(pos);
        } else if (strcmp(buf, "rxStbc") == 0) {
           ap_cfg->rxStbc = atoi(pos);
        } else if (strcmp(buf, "rifsModeAllowed") == 0) {
           ap_cfg->rifsModeAllowed = atoi(pos);
        } else if (strcmp(buf, "dualCtsProtection") == 0) {
           ap_cfg->dualCtsProtection = atoi(pos);
        } else if (strcmp(buf, "maxListenInterval") == 0) {
           ap_cfg->maxListenInterval = atoi(pos);
        }
        else {
           printf("Unknown parameter : %s on line = %d \n",buf,line);
           line++;
        }
    }
    fclose(fp);
    r = ioctl(fd, UNIFI_CFG, buffer);
    if (r < 0) {
        printf("UNIFI_CFG: failed\n");
        return r;
    } else {
        fprintf(stderr,"AP Configuration Set, Please Restart AP\n");
    }
    return 0;
}

int unifi_cfg_get_info(int fd, unifi_cfg_get_t type,
                       unsigned char *out_buffer)
{
    int r;
    unifi_cfg_command_t *pcmd = (unifi_cfg_command_t *)out_buffer;

    if (out_buffer == NULL) {
        return -1;
    }

    *pcmd = UNIFI_CFG_GET;
    *((unifi_cfg_get_t*)(out_buffer + sizeof(unifi_cfg_command_t))) = type;

    r = ioctl(fd, UNIFI_CFG, out_buffer);
    if (r < 0 ) {
        /* Do not print here in the operation is not permitted */
        /* This is done to avoid confusing prints for AP operation
           when builds are done seperately.. causing mismatching builds
        */
        if(r != -EPERM)
            printf("unifi_cfg_get_info: failed\n");
        return r;
    }

    return 0;
}

int unifi_cfg_set_strict_draft_n(int fd, CsrBool strict_draft_n)
{
    int r;
    unsigned char buffer[8];
    unifi_cfg_command_t *pcmd = (unifi_cfg_command_t *)buffer;

    *pcmd = UNIFI_CFG_STRICT_DRAFT_N;
    *((CsrBool*)(buffer + sizeof(unifi_cfg_command_t))) = strict_draft_n;

    r = ioctl(fd, UNIFI_CFG, buffer);
    if (r < 0) {
        printf("unifi_cfg_set_strict_draft_n: failed\n");
        return r;
    }

    return 0;
}

int unifi_cfg_set_enable_okc(int fd, CsrBool enable_okc)
{
    int r;
    unsigned char buffer[8];
    unifi_cfg_command_t *pcmd = (unifi_cfg_command_t *)buffer;

    *pcmd = UNIFI_CFG_ENABLE_OKC;
    *((CsrBool*)(buffer + sizeof(unifi_cfg_command_t))) = enable_okc;

    r = ioctl(fd, UNIFI_CFG, buffer);
    if (r < 0) {
        printf("unifi_cfg_set_enable_okc: failed\n");
        return r;
    }

    return 0;
}

#ifdef CSR_ONLY_DEBUG
/* Softmac native driver */
int unifi_cfg_add_softmac_template(int fd, const CsrUint8 *da, CsrUint16 pri, CsrUint8 hdr_len, const CsrUint8 *hdr_buf)
{
    int r;
    unsigned char buffer[sizeof(unifi_softmac_command_t)+sizeof(CsrUint8)+sizeof(unifiio_ma_packet_template_t)];
    unifi_softmac_command_t *pcmd = (unifi_softmac_command_t*)buffer;
    CsrUint8 *plen = (CsrUint8*)(pcmd+1);
    unifiio_ma_packet_template_t *pmpt = (unifiio_ma_packet_template_t*)(plen+1);

    *pcmd = UNIFI_SOFTMAC_MA_PACKET_TEMPLATE;
    *plen = sizeof(*pmpt);
    if (!da) {
        /* Wildcard if null */
        memset(pmpt->da, 0, sizeof(pmpt->da));
    }
    else {
        memcpy(pmpt->da, da, sizeof(pmpt->da));
    }
    pmpt->priority = pri;
    pmpt->template_len = hdr_len;
    pmpt->template_buf = hdr_buf;
#if 0
    fprintf(stderr,"ioctl 0x%08x %p\n\t", UNIFI_SOFTMAC_CFG, buffer);
    for(r=0; r<sizeof(buffer); r++)
        fprintf(stderr,"%02x ", buffer[r]);
    fprintf(stderr, "\n\t\t");
    for(r=0; r<hdr_len; r++)
        fprintf(stderr,"%02x ", hdr_buf[r]);
    fprintf(stderr, "\n");
#endif
    r = ioctl(fd, UNIFI_SOFTMAC_CFG, buffer);
    if (r < 0) {
        printf("unifi_cfg_add_softmac_template: failed\n");
        return r;
    }

    return 0;
}

int unifi_cfg_del_softmac_template(int fd, const CsrUint8 *da, CsrUint16 pri)
{
    /* Delete is add with no template */
    return unifi_cfg_add_softmac_template(fd,da,pri,0,0);
}

int unifi_cfg_set_softmac_filter(int fd, CsrUint8 precedence, CsrUint8 action, CsrUint8 filter_len, const CsrUint8* content_buf, const CsrUint8* mask_buf)
{
    int r;
    unsigned char buffer[sizeof(unifi_softmac_command_t)+sizeof(CsrUint8)+sizeof(unifiio_ma_packet_filter_t)];
    unifi_softmac_command_t *pcmd = (unifi_softmac_command_t*)buffer;
    CsrUint8 *plen = (CsrUint8*)(pcmd+1);
    unifiio_ma_packet_filter_t *pmpf = (unifiio_ma_packet_filter_t*)(plen+1);

    *pcmd = UNIFI_SOFTMAC_MA_PACKET_FILTER;
    *plen = sizeof(*pmpf);
    pmpf->precedence = precedence;
    pmpf->action = action;
    pmpf->offset = 0;
    while (filter_len && !*mask_buf)
    {
        /* Byte completely masked so remove from filter */
        pmpf->offset++;

        content_buf++;
        mask_buf++;
        filter_len--;
    }
    pmpf->filter_len = filter_len;
    pmpf->content_buf = content_buf;
    pmpf->mask_buf = mask_buf;

#if 0
    fprintf(stderr,"ioctl 0x%08x %p\n\t", UNIFI_SOFTMAC_CFG, buffer);
    for(r=0; r<sizeof(buffer); r++)
        fprintf(stderr,"%02x ", buffer[r]);
    fprintf(stderr, "\n%p:\t", content_buf);
    for(r=0; r<filter_len; r++)
        fprintf(stderr,"%02x ", content_buf[r]);
    fprintf(stderr, "\n%p:\t", mask_buf);
    for(r=0; r<filter_len; r++)
        fprintf(stderr,"%02x ", mask_buf[r]);
    fprintf(stderr, "\n");
#endif
    r = ioctl(fd, UNIFI_SOFTMAC_CFG, buffer);
    if (r < 0) {
        printf("unifi_cfg_set_softmac_filter: failed\n");
        return r;
    }

    return 0;
}
#endif /* CSR_ONLY_DEBUG */
